<copyright> DNEWS Server class.
    Written by <a href="mailto:tiggr@es.ele.tue.nl">Pieter J. Schoenmakers</a>

    Copyright &copy; 1996 Pieter J. Schoenmakers.
    All rights reserved.

    <id>$Id: Server.t,v 1.7 1996/11/25 17:26:28 tiggr Exp $</id>
    </copyright>

implementation class
Server: State, instance (DescriptorReadDelegate), stdio, Conditions
{
  <doc> Our news database.  </doc>
  public static Database database;
 
  <doc> The server's server port.  </doc>
  static ServerInetPort server_port;

  <doc> The name of this news system.  </doc>
  String system_name;

  <doc> The current (or most recent) server.  </doc>
  // CCC If this says `instance (id)', it is not properly handled (i.e. it
  // is expanded in the context of the user instead of this declaration).
  // Fri Aug 16 15:26:23 1996, tiggr@cobra.es.ele.tue.nl
  Server current_server;
}

OutputStream
  errorStream
{
  [err print ([Date now], " ")];
  = !current_server ? err : [current_server errorStream err];
}

<doc> Given the {(start, length)} range in the buffer {b} return a mapped
    collection containing all headers, stripped from whitspace.  If {nil}
    is returned, an invalid header field was envountered.  </doc>
MutableMapped
  headersFrom String b
        range (int, int) (start, length)
{
  /* Just to be safe.  */
  (start, length) = [b adjustRange (start, length)];

  /* Ensure that at least the last line is correctly terminated.  */
  if (length < 2 || b[length - 1] != '\n' || b[length - 2] != '\r')
    return nil;

  MutableDictionary headers = [MutableDictionary new];
  int i, end_pos = start + length, start_value;
  MutableByteString field;

  for (i = start; i < end_pos; i++)
    {
      byte c = b[i];

      if (c >= 'A' && c <= 'Z')
	c += ' ';

      // Should (implement and) use `isspace' and such.
      // Thu Aug 15 17:41:57 1996, tiggr@cobra.es.ele.tue.nl
      if (c == ' ' || c == '\t' || c == '\r' || c == '\n')
	return nil;
      if (c != ':')
	{
	  if (!field)
	    field = [MutableByteString new];
	  [field add c];
	}
      else
	{
	  MutableByteString value = [MutableByteString new];

	  /* A header with an empty name.  */
	  if (!field)
	    return nil;

	  /* Skip leading whitespace.  */
	  while (++i < end_pos)
	    {
	      c = b[i];

	      if (c == '\r' || c == '\n')
		{
		  /* This is an empty header field.  */
		  return nil;
		}
	      if (c != ' ' && c != '\t')
		break;
	    }

	  /* Collect multiple lines of this header.  */
	  for (;;)
	    {
	      while (i < end_pos)
		{
		  c = b[i++];

		  /* Bail out on bad termination.  */
		  if (c == '\r')
		    {
		      if (b[i] == '\n')
			break;
		      else
			return nil;
		    }
		  else if (c == '\n')
		    return nil;

		  [value add c];
		}

	      /* Break the loop if we're done with this header.  */
	      if (i + 1 == end_pos || (b[i + 1] != ' ' && b[i + 1] != '\t'))
		break;

	      /* Skip whitespace.  */
	      while (++i < end_pos)
		{
		  c = b[i];

		  if (c == '\r' || c == '\n')
		    {
		      /* This is at least an empty continuation line.  */
		      return nil;
		    }
		  if (c != ' ' && c != '\t')
		    break;
		}

	      [value add ' '];
	      [value add c];
	    }

	  [[[Server errorStream] print ("`", field, "': `", value, "'")] nl];
	  headers[field] = value;
	  field = nil;
	  start = i + 1;
	}
    }

  return headers;
}

<doc> Given the comma seperate list of newsgroups return an {Indexed}
    collection containing those newsgroups (as {Group} objects) which are
    valid.  </doc>
Indexed
  validNewsgroupsFrom String csl
{
  Indexed names = [csl componentsSeparatedBy ','];
  int i, n = [names length];
  MutableArray groups;

  for (i = 0; i < n; i++)
    {
      String name = names[i];
      int si, sn = [name length];
      int start = -1, the_end = -1;

      for (si = 0; si < sn; si++)
	{
	  char c = name[si];

	  if (c == ' ' || c == '\t')
	    {
	      if (start != -1 && the_end == -1)
		the_end = si;
	    }
	  else if (start == -1)
	    start = si;
	  else if (the_end != -1)
	    break;
	}

      if (si != sn || start == -1)
	[[[Server errorStream]
	  print ("ERROR invalid newsgroup `", name, "'")] nl];
      else
	{
	  if (start != 0 || the_end != -1)
	    {
	      if (the_end == -1)
		the_end = sn;
	      name = [name substring (start, the_end - start)];
	    }

	  name = [name downcase];
	  Group g = [Group named name];

	  if (!g)
	    [[[Server errorStream] print ("WARNING no group `", name, '\'')]
	      nl];
	  else
	    {
	      if (!groups)
		groups = [MutableObjectArray new];
	      [groups add g];
	    }	    
	}
    }

  = groups;
}

int
  main Array arguments
{
  system_name = "cobra.es.ele.tue.nl";

  bind ((error,
	 {
	   [[[Server errorStream] print ("server port allocation failed: ",
					 [condition message])] nl];
	   [Runtime exit 1];
	   nil;
	 }))
    server_port = [ServerInetPort with 1119 at nil];

  database = [Database connect "dnews" host: "viper.es.ele.tue.nl"];
  [Group initializeFrom database];

  [[RunLoop current] addDescriptorForRead server_port delegate self];
  [[RunLoop current] run];

  = 0;
}

void
  readEventOnDescriptor Descriptor d
{
  ConnectedInetPort client_port = [server_port accept];

  [[[Server errorStream]
    print ("connection from ", [[[[client_port peer] address] host] name])] nl];

  [[self alloc] init client_port];
}

end;

implementation instance
Server: instance (DescriptorReadDelegate), instance (DescriptorWriteDelegate)
{
  <doc> Our socket.  </doc>
  ConnectedInetPort client;

  <doc> Our state.  </doc>
  int state;

  <doc> The current group, if any.  </doc>
  Group group;

  <doc> The article pointer, or 0 if not set.  </doc>
  int article_pointer;

  <doc> The current article, if any.  </doc>
  Article article;

  <doc> Different article retrieval commands.  </doc>
  const AR_ARTICLE = 0;
  const AR_HEAD = 1;
  const AR_BODY = 2;
  const AR_STAT = 3;

  // Always having the descriptor open for read can be a problem if the
  // client issues commands too fast, without waiting for our replies.
  // Thu Aug 15 14:04:57 1996, tiggr@cobra.es.ele.tue.nl
  <doc> Possible states.  If bit 1 is set, we are writing asynchronously.
    </doc>
  const STATE_PLAIN = 0;
  const STATE_LIST = 1;

  <doc> The enumerator from which we LIST the groups.  </doc>
  Enumerator en;

  const STATE_POST = 2;

  <doc> The buffer into which the article is collected, in NNTP
    transmission form.  </doc>
  MutableByteString posting_buffer;

  <doc> The start of the body (i.e. the index of the second "\r\n" in the
    first occuring sequence "\r\n\r\n").  </doc>
  int body_start;
}

OutputStream
  errorStream OutputStream s
{
  = [s print ([[[[client peer] address] host] name], " ")];
}

void
  cleanup
{
  [[[Server errorStream] print "DEBUG [Server cleanup]"] nl];

  [[RunLoop current] removeReadDescriptor client];
  if (state & 1 != 0)
    [[RunLoop current] removeWriteDescriptor client];
  [client close];
  client = nil;
}

id
  init ConnectedInetPort port
{
  // The client stream should be buffered, but this suffers two problems:
  // (1) The BufferedStream does not yet handle bidirectional buffering,
  // though this could be solved by using two BufferedStreams, one for
  // input and one for output.  (2) The RunLoop does not yet handle
  // buffered streams; if buffered input is available, this will be
  // ignored and the RunLoop will wait on the descriptor.
  // Thu Aug 15 14:04:25 1996, tiggr@cobra.es.ele.tue.nl
  client = port;

  [[RunLoop current] addDescriptorForRead client delegate self];

  [[client print ("200 ", [Runtime hostname], " dnews server ready\r")] nl];

  = self;
}

<doc> Implementation of the LAST and NEXT commands.  </doc>
String
  shiftArticlePointer boolean forward
{
  if (!group)
    return "412 no newsgroup selected";
  if (!article_pointer)
    return "420 no current article has been selected";

  int first = [group first], last = [group last], ap = article_pointer;
  ObjectID gid = [group oid];
  AinG aing;

  for (;;)
    {
      if (forward && ++ap > last)
	return "421 no next article in this group";
      else if (!forward && --ap < first)
	return "422 no previous article in this group";

      aing = [AinG article ap inGroup gid inDatabase database];
      if (!!aing)
	break;
    }

  article_pointer = ap;

  [[client print ("223 ", article_pointer, " ", [aing message_id],
		  " article retrieved - request text separately\r")] nl];
}

<doc> Do the hard work for POST.  Return whatever to be replied to the
    client.  </doc>
String
  postArticle
{
  if (!body_start)
    return "441 posting failed (no body)";

  MutableMapped headers = [isa headersFrom posting_buffer
			       range (0, body_start)];
  if (!headers)
    return "441 posting failed (bad headers)";

  String h_newsgroups = headers["newsgroups"];
  String h_message_id = headers["message-id"];
  String h_subject = headers["subject"];
  String h_from = headers["from"];
  String h_path = headers["path"];
  MutableByteString prepend = [MutableByteString new];

  if (!h_from)
    return "441 posting failed (no `From:' header)";
  if (!h_subject)
    return "441 posting failed (no `Subject:' header)";
  if (!h_newsgroups)
    return "441 posting failed (no `Newsgroups:' header)";

  if (!h_message_id)
    {
      h_message_id = [[MutableByteString withCapacity 30]
		       print ('<', [Date timeIntervalSinceReferenceDate],
			      '@', [Runtime hostname], '>')];
      headers["message-id"] = h_message_id;
      [prepend print ("Message-ID: ", h_message_id, "\r\n")];
    }

  if (!headers["date"])
    {
      String date = [[MutableByteString new] print [Date now]];
      headers["date"] = date;
      [prepend print ("Date: ", date, "\r\n")];
    }

  if (!h_path)
    [prepend print ("Path: ", system_name, "\r\n")];
  else
    {
      int s, l;

      (s, l) = [h_path rangeOfString system_name];
      if (l > 0
	  && (!s || h_path[s - 1] == '!')
	  && (s + l == [h_path length] || h_path[s + l] == '!'))
	return "437 article rejected (found me in path)";
    }

  Indexed groups = [isa validNewsgroupsFrom h_newsgroups];

  if (!groups)
    return "441 posting failed (no valid/known newsgroups)";

  [database begin];

  [prepend print posting_buffer];
  Article a = [Article createWithContents prepend
		       messageID [h_message_id uniqueString]
		       inDatabase database];
  int i, n = [groups length];

  for (i = 0; i < n; i++)
    [groups[i] aingForArticle a];

  [database commit];

  = "240 article posted ok";
}

<doc> Do the work for the ARTICLE command.  Upon failure, return a non-nil
    string describing what went wrong.  If {art} is nil, the current
    article as set through the article or article pointer is handled.
    </doc>
String
  retrieve int how
   article UniqueString art
{
  boolean b;
  int v;

  (v, b, ) = [art integerValue (0, -1) allowSign: NO allowCBases: NO];
  if (b)
    if (!group)
      return "412 no newsgroup selected";
    else
      {
	AinG aing = [AinG article v inGroup [group oid] inDatabase database];
	if (!aing)
	  return "423 no such article number in this group";
	art = [aing message_id];
      }

  Article at = [Article withMessageID art inDatabase database];
  if (!at)
    if (b)
      return "423 no such article number in this group";
    else
      return "430 no such article found";

  article = at;
  if (b)
    article_pointer = v;

  ByteArray contents = [article contents];

  if (!contents)
    return "503 program fault - command not performed";

  if (how == AR_ARTICLE)
    {
      [[client print ("220 0 ", [article message_id],
	       " article retrieved - head and body follow")] nl];
      [[client print contents] nl];
    }
  else
    {
      [[[Server errorStream]
	print ("ERROR bad way ", how, " for article retrieve")] nl];
      return "503 program fault - command not performed";
    }

  = nil;
}

<doc> The heart of user interaction.  </doc>
void
  readEventOnDescriptor Descriptor d
{
  current_server = self;

  if (state == STATE_PLAIN)
    {
      // This isn't really suitable for asynchronous operation...
      // Wed Aug 14 11:54:23 1996, tiggr@cobra.es.ele.tue.nl
      MutableByteString command = [client readLine];

      if (!command)
	[self cleanup];
      else
	{
	  int l = [command length];
	  String bad;

	  if (!l || command[l - 1] != '\r')
	    bad = "500 command not recognized";
	  else
	    {
	      [command resize --l];

	      Array a = [command componentsSeparatedBy ' ' excludeEmpty: YES];
	      int argc = [a length];

	      if (argc > 0 && [a[0] length] > 0)
		{
		  String cmd = [a[0] downcase];
		  byte first = cmd[0];

		  if (["es" equal cmd])
		    if (argc == 1)
		      {
			[[[Server errorStream] print "EXIT SERVER"] nl];
			[Runtime exit 0];
		      }
		    else
		      bad = "500 command not recognized";
		  else if (["article" equal cmd])
		    if (argc == 2)
		      bad = [self retrieve AR_ARTICLE
				  article [a[1] uniqueString]];
		    else
		      bad = "501 usage: ARTICLE (n|<id>)";
		  else if (["group" equal cmd])
		    if (argc == 2)
		      {
			Group g = [Group named a[1]];
			if (!g)
			  bad = "411 no such news group";
			else
			  {
			    group = g;
			    article = nil;

			    int f = [group first];
			    int l = [group last];
			    int n = l < f ? 0 : 1 + l - f;

			    article_pointer = n ? [group first] : 0;

			    [[client print ("211 ", n, " ", f, " ", l, " ",
					    [group name], " group selected\r")]
			     nl];
			  }
		      }
		    else
		      bad = "501 usage: GROUP <name>";
		  else if (["help" equal cmd])
		    if (argc == 1)
		      [[client print
			       "100 help text follows\r
  ARTICLE n|i retrieve article with number n or message-id i\r
  GROUP name  select the group with the <name>\r
  HELP        output this message\r
  LAST        select previous message in current group\r
  LIST        list the known groups\r
  NEXT        select next message in current group\r
  POST        express some view to the world\r
  QUIT        close connection\r
  SLAVE       anachronism\r
.\r"] nl];
		    else
		      bad = "501 usage: HELP";
		  else if (["last" equal cmd])
		    if (argc == 1)
		      bad = [self shiftArticlePointer NO];
		    else
		      bad = "501 usage: LAST";
		  else if (["list" equal cmd])
		    if (argc == 1)
		      {
			[[client print "215 list of newsgroups follows\r"] nl];
			state = STATE_LIST;
			en = [[Group groups] enumerator];
			// Maybe we should remove ourselves for the read
			// event...
			[[RunLoop current] addDescriptorForWrite client
					   delegate self];
		      }
		    else
		      bad = "501 usage: LIST";
		  else if (["next" equal cmd])
		    if (argc == 1)
		      bad = [self shiftArticlePointer YES];
		    else
		      bad = "501 usage: NEXT";
		  else if (["post" equal cmd])
		    if (argc == 1)
		      {
			[[client print "340 send article to be posted\r"] nl];
			state = STATE_POST;

			if (!posting_buffer)
			  posting_buffer = [MutableByteString withCapacity
							      8192];
			else
			  [posting_buffer resize 0];
			body_start = 0;
		      }
		    else
		      bad = "501 usage: POST";
		  else if (["quit" equal cmd])
		    if (argc == 1)
		      {
			[[client print "205 closing connection - goodbye!\r"]
			 nl];
			[self cleanup];
		      }
		    else
		      bad = "501 usage: QUIT";
		  else if (["slave" equal cmd])
		    if (argc == 1)
		      [[client print "202 perceived world view noted\r"] nl];
		    else
		      bad = "501 usage: SLAVE";
		  else
		    bad = "500 command not recognized";
		}
	      else
		bad = "500 command not recognized";
	    }

	  if (bad != nil)
	    [client print (bad, "\r\n")];
	}
    }
  else if (state == STATE_POST)
    {
      // This code loses if more stuff is sent without waiting for the
      // reply.  Then the \r\n.\r\n will be somewhere in between and will
      // not be noticed.  We should actually read a character at a time...
      // Thu Aug 15 16:55:07 1996, tiggr@cobra.es.ele.tue.nl
      int buf_len = [posting_buffer length];
      int read = [client readBytes 8192 into posting_buffer];
      int start, length;

      if (!read)
	{
	  [self cleanup];
	  return;
	}

      if (body_start == 0)
	{
	  (start, length) = [posting_buffer rangeOfString "\r\n\r\n"
			      range: (buf_len > 3 ? buf_len - 4 : 0, -1)];
	  if (length > 0)
	    body_start = start + 2;
	}

      buf_len += read;

      (start, length) = [posting_buffer rangeOfString "\r\n.\r\n"
					range: (buf_len - 5, 5)];
      if (length > 0)
	{
	  [[client print ([self postArticle], '\r')] nl];
	  state = STATE_PLAIN;
	}
    }
  else
    {
      [[[Server errorStream]
	print ("ERROR bad state ", state, " for read event")] nl];
      [self cleanup];
    }
  current_server = nil;
}

<doc> The second heart of user interaction.  </doc>
void
  writeEventOnDescriptor Descriptor d
{
  current_server = self;
  if (state == STATE_LIST)
    {
      boolean b;
      Group g;

      (b, g) = [en next];

      if (!b)
	{
	  [client print ".\r\n"];
	  state = STATE_PLAIN;
	  [[RunLoop current] removeWriteDescriptor client];
	}
      else
	[[client print ([g name], " ", [g last], " ", [g first], " y\r")] nl];
    }
  else
    {
      [[[Server errorStream]
	print ("ERROR bad state ", state, " for write event")] nl];
      [self cleanup];
    }
  current_server = nil;
}

end;
