/* C routines for dnews.
   Written by Pieter J. Schoenmakers <tiggr@es.ele.tue.nl>

   Copyright (C) 1996 Pieter J. Schoenmakers.
   All rights reserved.

   $Id: glue.c,v 1.4 1996/10/20 11:56:54 tiggr Exp $  */

#include <libpq-fe.h>
#include <libpq/libpq-fs.h>
#include "dnews-r.h"
#include <tom/util.h>

void
i_dnews_Postgres_v_close (tom_object self, selector cmd)
{
  struct _es_i_dnews_Postgres *this
    = trt_ext_address (self, _ei_i_dnews_Postgres);

  PQfinish (this->connection);
  this->connection = 0;
}

void
i_dnews_Postgres_v_dealloc (tom_object self, selector cmd)
{
  struct _es_i_dnews_Postgres *this
    = trt_ext_address (self, _ei_i_dnews_Postgres);

  if (this->result)
    {
      PQclear (this->result);
      this->result = NULL;
    }

  if (this->connection)
    TRT_SEND (, self, SEL (v_close));
}

tom_object
i_dnews_Postgres_r_init__rr_ (tom_object self, selector cmd,
			      tom_object database, tom_object hostname)
{
  char *dbn = c_string_with_tom_string (database);
  char *hn = c_string_with_tom_string (hostname);
  PGconn *db = PQsetdb (hn, NULL, NULL, NULL, dbn);
  struct _es_i_dnews_Postgres *this
    = trt_ext_address (self, _ei_i_dnews_Postgres);

  if (PQstatus (db) != CONNECTION_OK)
    abort ();

  this->hostname = (void *) hostname;
  this->name = (void *) database;
  this->connection = db;

  xfree (dbn);
  xfree (hn);

  return self;
}

void
i_dnews_Postgres_v_exec_r (tom_object self, selector cmd, tom_object query)
{
  struct _es_i_dnews_Postgres *this
    = trt_ext_address (self, _ei_i_dnews_Postgres);
  char *s = c_string_with_tom_string (query);

  if (this->result)
    PQclear (this->result);
  this->result = PQexec (this->connection, s);
  if (!this->result)
    TRT_RAISE (self, c_tom_Conditions_error,
	       byte_string_with_c_string (((PGconn *) this->connection)
					  ->errorMessage));

  this->next_row = 0;
}

tom_int
i_dnews_Postgres_i_numRows (tom_object self, selector cmd)
{
  struct _es_i_dnews_Postgres *this
    = trt_ext_address (self, _ei_i_dnews_Postgres);

  return this->result ? PQntuples (this->result) : 0;
}

tom_object
i_dnews_Postgres_r_attributes (tom_object self, selector cmd)
{
  struct _es_i_dnews_Postgres *this
    = trt_ext_address (self, _ei_i_dnews_Postgres);
  tom_object attr;
  tom_int i, n;

  if (!this->result || !PQntuples (this->result))
    return 0;

  n = PQnfields (this->result);
  attr = TRT_SEND ((pointer_imp), _mr_c_tom_MutableArray,
		   SEL (r_withCapacity_i), n);

  for (i = 0; i < n; i++)
    TRT_SEND ((void_imp), attr, SEL (v_add_r),
	      byte_string_with_c_string (PQfname (this->result, i)));

  return attr;
}

tom_object
i_dnews_Postgres_r_nextRow (tom_object self, selector cmd)
{
  struct _es_i_dnews_Postgres *this
    = trt_ext_address (self, _ei_i_dnews_Postgres);
  tom_object values;
  tom_int i, n;

  if (!this->result || this->next_row >= PQntuples (this->result))
    return 0;

  n = PQnfields (this->result);
  values = TRT_SEND ((pointer_imp), _mr_c_tom_MutableArray,
		     SEL (r_withCapacity_i), n);

  for (i = 0; i < n; i++)
    TRT_SEND ((void_imp), values, SEL (v_add_r),
	      byte_string_with_c_string (PQgetvalue (this->result,
						     this->next_row, i)));
  this->next_row++;

  return values;
}

tom_object
i_dnews_Postgres_r_fetchAttributes_r (tom_object self, selector cmd,
				      tom_object attributes)
{
  struct _es_i_dnews_Postgres *this
    = trt_ext_address (self, _ei_i_dnews_Postgres);
  tom_object values;
  int j, nfields;
  tom_int i, n;

  if (!this->result || this->next_row >= PQntuples (this->result))
    return 0;

  values = TRT_SEND ((pointer_imp), _mr_c_tom_MutableDictionary, SEL (r_new));
  nfields = PQnfields (this->result);
  n = TRT_SEND (, attributes, SEL (i_length));

  for (i = 0; i < n; i++)
    {
      tom_object attribute;
      tom_int l;
      char *s;

      attribute = TRT_SEND ((pointer_imp), attributes, SEL (r_at_i), i);
      s = TRT_SEND ((pointer_imp), attribute, SEL (_pi__byteStringContents),
		    &l);

      for (j = 0; j < nfields; j++)
	if (!memcmp (PQfname (this->result, j), s, l))
	  {
	    TRT_SEND ((void_imp), values, SEL (v_set_r_at_r),
		      byte_string_with_c_string
		        (PQgetvalue (this->result, this->next_row, j)),
		      attribute);
	    break;
	  }
    }

  this->next_row++;

  return values;
}

tom_object
i_dnews_Postgres_r_largeObjectCreate_r (tom_object self, selector cmd,
					tom_object contents)
{
  struct _es_i_dnews_Postgres *this
    = trt_ext_address (self, _ei_i_dnews_Postgres);
  Oid lo = lo_creat (this->connection, INV_READ | INV_WRITE);
  tom_int l;
  char *s;
  int od;

  if (!OidIsValid (lo))
    /* XXX Calling abort isn't valid either.  */
    abort ();

  s = TRT_SEND ((pointer_imp), contents, SEL (_pi__byteStringContents), &l);

  od = lo_open (this->connection, lo, INV_WRITE);
  if (od < 0)
    abort ();

  if (l != lo_write (this->connection, od, s, l))
    abort ();

  lo_close (this->connection, od);
  return TRT_SEND ((pointer_imp), _mr_c_tom_IntNumber, SEL (r_with_i), lo);
}

tom_object
i_dnews_Postgres_r_largeObjectRetrieve_r (tom_object self, selector cmd,
					  tom_object identity)
{
  struct _es_i_dnews_Postgres *this
    = trt_ext_address (self, _ei_i_dnews_Postgres);
  Oid lo = TRT_SEND (, identity, SEL (i_intValue));
  int od, i, len, cap;
  char *s;

  od = lo_open (this->connection, lo, INV_READ);
  if (od < 0)
    abort ();

  /* XXX Hohum, from the sources it seems anything different from L_SET is
     not supported.  Oh well... we'll reallocate as we go...  */
  len = cap = 0;
  s = 0;

  for (;;)
    {
      if (len == cap)
	{
	  cap += 1024;
	  s = xrealloc (s, cap);
	}

      i = lo_read (this->connection, od, s + len, cap - len);
      if (i > 0)
	len += i;
      else if (i == 0)
	break;
      else
	{
	  lo_close (this->connection, od);

	  return TRT_SIGNAL (self, c_tom_Conditions_error,
			     byte_string_with_c_string
			     (((PGconn *) this->connection) ->errorMessage));
	}
    }

  s = xrealloc (s, len);

  lo_close (this->connection, od);

  return byte_string_with_string (s, len);
}

tom_object
i_dnews_Postgres_r_resultOfInsert (tom_object self, selector cmd)
{
  struct _es_i_dnews_Postgres *this
    = trt_ext_address (self, _ei_i_dnews_Postgres);
  char *s = PQoidStatus (this->result);

  if (!*s)
    TRT_RAISE (self, c_tom_Conditions_error,
	       byte_string_with_c_string ("previous command was not an insert"
					  " or not successful"));
    
  return TRT_SEND ((pointer_imp), _mr_c_tom_IntNumber,
		   SEL (r_with_i), atoi (s));
}
